/*
 * Map File Generator - Original version 1
 *
 * This program reads a DDS listing file, and generates a .MAP file
 * containing information about file names, line number/address mapping,
 * and basic global symbol information.
 *
 * Copyright 1994-2000 Dave Dunfield
 * Freely distributable.
 *
 * Permission is granted to modify and distribute this program without
 * payment or royalty, as long as the original copyright notices are not
 * removed. ** Portability notes are identified by **

 * The format of this example .MAP file is:

	(Example ASCII MAP file)		<- Header
	Files=n							<- Number of filenames to follow
	filenames (1 per line)			<- Name of each file
	Records=n						<- Number of address records to follow
	Lxxxx f l [s symbol]			<- Line address record *
	Exxxx s symbol					<- Equate address record *

 * Line record format:
	L			<- Indicates this is a LINE record
	xxxx		<- Hexidecimal code address
	f			<- Filename index (1 = 1st file, 2 = 2nd file...)
	l			<- Line number in file where this address occurs
	[s]			<- Map segment where symbol is defined
	[symbol]	<- Symbol name (if present at this address)

 * Equate record format
	E			<- Indicates this is an EQUATE record
	xxxx		<- Hexidecimal code address
	s			<- Map segment where symbol is defined
	symbol		<- Name of symbol occuring at this address

 * LINE records and EQUATE records can occur in any order.
 */
#include <stdio.h>

#define	LINE_SIZE	250		/* Maximum size of input lines */
#define	MAX_FILES	100		/* Maximum number of files to handle */

/* Offsets into listing lines */
#define	OPCODE	7			/* Opcode bytes begin here */
#define	SOURCE	32			/* Source text begin here */

/* Information parsed from the input file */
unsigned addr, cline = 0, last_cline = 0, symaddr, lst_line = 0,
	current_file = 0, mapseg = 0;
char symbol[25], filenames[MAX_FILES][65], pass = 0;

/* Counters */
unsigned file_count = 0, num_entry = 0, num_equ = 0, num_sym = 0;

/* Command line parameters */
char xfile = -1, conly = 0, dup = 0, equ = -1, list = 0, quiet = 0, ucase = 0;

/* Input and Output file pointers */
FILE *fpi, *fpo;

/*
 * Main program - Parse parameters & process listing file
 */
main(argc, argv)
	int argc;
	char *argv[];
{
	int i;
	char buffer[LINE_SIZE+1], *ptr, c;

	/* Parse for command line options */
	for(i=2; i < argc; ++i) {
		if(*(ptr = argv[i]) == '-') {		/* Enable switch */
			while(*++ptr) switch(toupper(*ptr)) {
				case 'A' : xfile = 0;	break;	/* Map Assembly only */
				case 'C' : conly = -1;	break;	/* Map 'C' only */
				case 'D' : dup = -1;	break;	/* Allow duplicate lines */
				case 'E' : equ = 0;		break;	/* Inhibit EQU symbols */
				case 'L' : list = -1;	break;	/* Generate debug listing */
				case 'Q' : quiet = -1;	break;	/* Quite mode (no output) */
				case 'U' : ucase = -1;	break;	/* Uppercase everything */
				default: goto badopt; }
			continue; }
	badopt:
		printf("Invalid option: %s\n", argv[i]);
		exit(-1); }

	if(!quiet)
		printf("Map file generator, version 1.1\n");

	if(argc < 2) {
		printf("\nUse: MAPGEN <filename> [-acdelqu]\n\nCopyright 1994-2000 Dave Dunfield\nFreely distributable.\n");
/*
 * Please provide indication of MAP file supported below
 */
		printf("\nModified by <no-one> to support: ASCII Example\n");
		exit(-1); }

	/* Parse filename & extension from passed path etc. */
	copy_parse(ptr = buffer, argv[1]);
	while(c = *ptr) {
		if(c == '.')
			goto noext;
		++ptr; }
		copy_parse(ptr, ".LST");
noext:
/*
 * Open input and output files
 *
 ** Most 'C' compilers do not support MICRO-Cs 'v'erbose and 'q'uit
 ** options on fopen(), you may have to check for NULL return and
 ** issue error messages before exiting.
 **
 ** If your output .MAP file is in BINARY format, don't forget to
 ** add the 'b'inary option to the second fopen() statement.
 */
	copy_parse(filenames[file_count++], buffer);
	fpi = fopen(buffer, "rvq");
	copy_parse(ptr, ".MAP");
	fpo = fopen(buffer, "wvq");

	/*****************************************************/
	/* Output any pre-amble needed in the .MAP file here */
	/*****************************************************/
	fprintf(fpo, "(Example ASCII MAP file)\n");

	/* Pass#1 - Read listing & determine # files and entries */
	*symbol = 0;
	while(read_line(buffer)) {
		if(current_file) {		/* At 'C' level */
			if(cline) {
				++num_entry;
				*symbol = cline = 0; } }
		else if(!conly) {
			++num_entry;
			*symbol = 0; } }

	/**********************************************/
	/* Write number of files & table of filenames */
	/**********************************************/
	fprintf(fpo, "Files=%u\n", file_count);
	for(i=0; i < file_count; ++i) {
		fprintf(fpo, "%s\n", ptr = filenames[i]);
		if(list)
			printf("File: %s\n", ptr); }

	/***********************************/
	/* Write number of address records */
	/***********************************/
	fprintf(fpo, "Records=%u\n", num_entry + num_equ);

	/* Pass#2 - Read listing and generate address entries */
	rewind(fpi);
	current_file = cline = lst_line = mapseg = 0;
	pass = equ;

	while(read_line(buffer)) {
		if(current_file) {		/* At 'C' level */
			if(cline) {
				make_entry();
				cline = 0; } }
		else
			make_entry(); }

	fclose(fpo);
	fclose(fpi);

	if(!quiet)
		printf("%u files, %u address entries (%u with symbols), %u equ symbols\n",
		file_count, num_entry, num_sym, num_equ);
}

/*
 * Test for a valid symbol name
 */
issymbol(ptr)
	char *ptr;
{
	char c;
	c = *ptr;
	do {
		if( ((c < 'a') || (c > 'z'))
		&&	((c < 'A') || (c > 'Z'))
		&&	((c < '0') || (c > '9'))
		&&	(c != '_') )
			return 0; }
	while(!isspace(c = *++ptr));
	return 1;
}

/*
 * Test for string1 beginning with string2
 */
strbeg(str1, str2)
	char *str1, *str2;
{
	while(*str2)
		if(*str1++ != *str2++)
			return 0;
	return 1;
}

/*
 * Copy a symbol, but only up to a delimiter
 */
copy_parse(dest, source)
	char *dest, *source;
{
	if(ucase) {
		while(*source && !isspace(*source))
			*dest++ = toupper(*source++); }
	else {
		while(*source && !isspace(*source))
			*dest++ = *source++; }
	*dest = 0;
}


/*
 * Get a number in specified base with digit limit
 */
get_number(ptr, base, digits)
	char *ptr;
	unsigned base, digits;
{
	unsigned value, c;

	value = 0;
	while(isspace(*ptr))
		++ptr;
	do {
		if(isdigit(c = *ptr))			/* convert numeric digits */
			c -= '0';
		else if(c >= 'a')				/* convert lower case alphabetics */
			c -= ('a' - 10);
		else if(c >= 'A')				/* convert upper case alphabetics */
			c -= ('A' - 10);
		else							/* not a valid "digit" */
			break;
		if(c >= base)					/* outside of base */
			break;
		value = (value * base) + c;		/* include in total */
		++ptr; }
	while(--digits);					/* enforce maximum digits */
	return value;
}

/*
 * Read a line from the input file, and return it only if it is
 * a valid code associated address
 */
read_line(buffer)
	char *buffer;
{
	int i;
	char *ptr;

top:
	if(!fgets(buffer, LINE_SIZE, fpi))			/* End of input */
		return 0;
	++lst_line;
	if(*buffer < ' ') goto top;					/* Blank line */
	if(strbeg(buffer, "DUNFIELD")) goto top;	/* Header line */
	if(strbeg(buffer, "SYMBOL")) return 0;		/* End of listing */

	if(*(buffer+SOURCE) == '*') {				/* Comment line */
		if(*(buffer+(SOURCE+1)) == '#') {		/* Directive */
			if(strbeg(buffer+(SOURCE+2), "file ") && xfile) { /* Filename */
				copy_parse(ptr = filenames[file_count], buffer+(SOURCE+7));
				for(i=0; i < file_count; ++i)
					if(!strcmp(filenames[i], ptr)) {
						current_file = i;
						goto top; }
				current_file = file_count++; }
			else if(strbeg(buffer+(SOURCE+2), "map"))		/* Map segment */
				mapseg = atoi(buffer+(SOURCE+5));
			goto top; }
		if(isdigit(*(buffer+(SOURCE+1)))) {		/* 'C' line number */
			i = get_number(buffer+(SOURCE+1), 10, 0);
			if((i != last_cline) || dup)
				cline = last_cline = i;
			goto top; }
		/* Must be an ASM comment - reset to outer level */
		current_file = 0;
		goto top; }

	addr = get_number(buffer, 16, 4);			/* Get address */

	if(addr != symaddr)							/* We've moved off address */
		make_equate();

	if(issymbol(buffer+SOURCE)) {				/* Symbol on line */
		make_equate();
		copy_parse(symbol, buffer+SOURCE);
		symaddr = addr; }

	if(*(buffer+OPCODE) == ' ')					/* No opcode, don't save */
		goto top;

	return -1;
}

/*****************************************/
/* Make an address entry in the MAP file */
/*****************************************/
make_entry()
{
	unsigned line;

	if(current_file)			/* 'C' source file active */
		line = cline;
	else {						/* Index to listings file */
		if(conly)
			return;
		line = lst_line; }

	if(list)	/* Generate debug listing */
		printf("%04x F:%-2u  L:%-5u S:%u %s\n", addr, current_file+1,
		line, mapseg, symbol);

/* Output line record for MAP file here! */
	fprintf(fpo, "L%04x %u %u", addr, current_file+1, line);
	if(*symbol) {
		++num_sym;
		fprintf(fpo, " %u %s", mapseg, symbol);
		*symbol = 0; }
	putc('\n', fpo);
}

/************************************/
/* Make an "EQU" entry for a symbol */
/************************************/
make_equate()
{
	if(*symbol) {
		if(pass) {
			if(list)	/* Generate debug listing */
				printf("%04x F:%-2u  L:%-5u S:%u %s\n", addr, 0, 0,
					mapseg, symbol);

			/* Output EQU record to MAP file here! */
			fprintf(fpo, "E%04x %u %s\n", symaddr, mapseg, symbol);
			}

		else if(equ)
			++num_equ; }
	*symbol = 0;
}
